from typing import Callable, ParamSpec, TypeVar
from fastapi import APIRouter as FastApiRouter, Depends

T = TypeVar("T")
P = ParamSpec("P")


class ApiRouter(FastApiRouter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def set_app_dependency(self, app):
        self._app = app

    def _inject_app_dependency(self, endpoint: Callable[P, T]) -> Callable[P, T]:
        if not self._app:
            raise RuntimeError("App dependency not set.?")

        original_deps = getattr(endpoint, "dependencies", [])

        if not any(dep.dependency == self._app for dep in original_deps):
            endpoint.dependencies = [Depends(self._app)] + original_deps

        return endpoint

    def get(self, path: str, **kwargs):
        def decorator(endpoint: Callable[P, T]) -> Callable[P, T]:
            wrapped_endpoint = self._inject_app_dependency(endpoint)
            return super().get(path, **kwargs)(wrapped_endpoint)
        return decorator

    def post(self, path: str, **kwargs):
        def decorator(endpoint: Callable[P, T]) -> Callable[P, T]:
            wrapped_endpoint = self._inject_app_dependency(endpoint)
            return super().get(path, **kwargs)(wrapped_endpoint)
        return decorator

    def put(self, path: str, **kwargs):
        def decorator(endpoint: Callable[P, T]) -> Callable[P, T]:
            wrapped_endpoint = self._inject_app_dependency(endpoint)
            return super().get(path, **kwargs)(wrapped_endpoint)
        return decorator

    def delete(self, path: str, **kwargs):
        def decorator(endpoint: Callable[P, T]) -> Callable[P, T]:
            wrapped_endpoint = self._inject_app_dependency(endpoint)
            return super().get(path, **kwargs)(wrapped_endpoint)
        return decorator


from pydantic import PostgresDsn, field_validator
from pydantic_settings import BaseSettings as PydanticBaseSettings, SettingsConfigDict


class BaseSettings(PydanticBaseSettings):

    model_config = SettingsConfigDict(env_file=".env")

    database_host: str = "localhost"
    database_port: str = "5432"
    database_name: str
    database_user: str
    database_password: str
    database_url: PostgresDsn | None = None

    @field_validator("database_url")
    @classmethod
    def assemble_database_url(cls, v: PostgresDsn | None, values):
        return PostgresDsn.build(
            scheme="postgresql+psycopg",
            username=values.data.get("database_user"),
            password=values.data.get("database_password"),
            host=values.data.get("database_host"),
            port=int(values.data.get("database_port")),
            path=values.data.get("database_name"),
        )



from typing import Generator
from fastapi import FastAPI
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, Session

from nephyx.core.settings import BaseSettings
from nephyx.router import ApiRouter


class NephyxApi:
    def __init__(self, settings_cls: type[BaseSettings] = BaseSettings):
        self.settings = settings_cls()
        self.fastapi_app = FastAPI()

        self._setup_database()

    def _setup_database(self):
        self.db_engine = create_engine(self.settings.database_url)
        self._session_factory = sessionmaker(autocommit=False, autoflush=False, bind=self.db_engine)

    def include_router(self, router: ApiRouter):
        self.fastapi_app.include_router(router)
        router.set_app_dependency(lambda: self)

    def get_db(self) -> Generator[Session, None, None]:
        db = self._session_factory()
        try:
            yield db
        finally:
            db.close()



from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel


class BaseSchema(BaseModel):
    model_config = ConfigDict(
        from_attributes=True,
        coerce_numbers_to_str=False,
        alias_generator=to_camel,
        populate_by_name=True,
    )
